      program rootf
*
*  This is a polynomial root finder program.
*  the polynomials must have real coefficients.
*  The method applied is a modification of bairstow's 
*  i.e. division by a quadratic polynomial x**2 +p1*x +p2
*  determining p1 & p2 to be such that the remainder is vanishing.
*  Then a deflation is applied by synthetic division to reduce
*  the polynomial degree by two, and the procedure is repeated.
*  A check is always made if the degree is reduced to either
*  one or two to use special finishing procedure.
*
      implicit double precision (a-h,o-z)
      common /toroot/ a(0:100), b(0:100), kco
      complex root(100), disc
      parameter ( n = 2, m = 2 )
      dimension icode(4), xp(2), xll(2), xrl(2), ixat(2)
      character line*78
      data icode / 1, 0, 0, 0 /
*
      do 100 i=1,78
         line(i:i) = '='
100   continue
*
*  Read in, tolerance, polynomial degree, polynomial coefficients
      write (*,*) 'Enter a termination tolerance: '
      read (*,*) eps
      write (*,*) 'Enter the degree of the polynomial:'
      read (*,*) kco
      write (*,*) 'Enter the polynomial coefficients:( a(0),a(1),...)'
      read (*,*) (a(i),i=0,kco)
*
*  Type the above input for reference.
      write (*,'(a)') line
      write (*,*) '    Running with the following input  '
      write (*,*) 'Tolerance: ', eps
      write (*,*) 'Degree:    ', kco
      write (*,*)'        Coefficients  a(i), i=0,1,... '
      write (*,*) (a(i),i=0,kco)
      write (*,'(a)') line
      nor = 0
30    continue
*  Check if the degree is one, and if so solve a linear equation.
      if (kco.eq.1) then
         write (*,*) ' The roots of the polynomial are:'
         nor = nor + 1
         root(nor) = cmplx(-a(0)/a(1),0)
         do 1 i=1,nor
            write (*,19) i, root(i)
1        continue
19       format (2x,'root # ',i4,t17,'(',g14.7,',  ',g14.7,')')
         stop
      end if
* Check if the degree is two, and if so solve a quadratic equation.
      if (kco.eq.2) then
        write (*,*) ' The roots of the polynomial are:'
        nor = nor + 1
        disc = cmplx(a(1)**2 - 4*a(2)*a(0),0.d0)
        root(nor) = (-a(1)+csqrt(disc))/2./a(2)
        nor = nor + 1
        root(nor) = (-a(1)-csqrt(disc))/2./a(2)
        do 2 i=1,nor
           write (*,19) i, root(i)
2       continue
        stop
      end if
*
* Initialize randomly in (-1,1)
      p1 = 2*ranm()-1
      p2 = 2*ranm()-1
      xp(1) = p1
      xp(2) = p2
*
*  Minimize the remainder down to zero !!!
      call optima(n,m,xp,xv,xll,xrl,ixat,icode,
     &            'in.dat','/dev/null',grms,nf,ng,nhe,nja)
      if (sqrt(xv).le.eps) then
        p1 = xp(1)
        p2 = xp(2)
        disc = cmplx( p1**2-4*p2, 0.d0)
        nor = nor + 1
        root(nor)=(-p1 + csqrt(disc))/2
        nor = nor + 1
        root(nor)=(-p1 - csqrt(disc))/2
        call syndiv(a,kco,p1,p2,b,r,q)
        kco = kco - 2
        do 3 i = 0,kco
           a(i) = b(i)
3       continue
      end if
      go to 30
      end
*  ---------------------------------------------------------------------
      subroutine syndiv ( a, n, p1, p2, b, r, q )
*  ---------------------------------------------------------------------
      implicit double precision (a-h,o-z)
*
*  Given a polynomial: a(n)*x**n + ...+a(1)*x +a(0)
*  and a quadratic: x**2 + p1*x + p2
*  calculate the coefficients of the quotient polynomial: b(k),k=0,1,...,n-2
*  and the remainder terms r and q.
*
      dimension a(0:*), b(0:*)
*
      b(n-2) = a(n)
      b(n-3) = a(n-1)-p1*b(n-2)
      do 1 k=n-2,2,-1
         b(k-2) = a(k)-p1*b(k-1)-p2*b(k)
1     continue
      r = a(1) - p1*b(0)-p2*b(1)
      q = a(0) - p2*b(0)
      end
*  ---------------------------------------------------------------------
      subroutine synder (a, n, p1, p2, b, r, q, b1, b2, r1, r2, q1, q2)
*  ---------------------------------------------------------------------
      implicit double precision (a-h,o-z)
*
*  Given a polynomial: a(n)*x**n + ...+a(1)*x +a(0)
*  and a quadratic: x**2 + p1*x + p2
*  calculate the coefficients of the quotient polynomial: b(k),k=0,1,...,n-2
*  and the remainder terms r and q.
*  Calculate the derivatives: b1(k),r1,q1 of b(k),r and q with respect to p1, 
*  and the derivatives: b2(k),r2,q2 of b(k),r and q with respect to p2.
*
      dimension a(0:*),b(0:*),b1(0:*),b2(0:*)
*
      b(n-2) = a(n)
      b(n-3) = a(n-1) - p1*b(n-2)
      do 1 k=n-2,2,-1
         b(k-2) = a(k) - p1*b(k-1) - p2*b(k)
1     continue
      r = a(1) - p1*b(0)-p2*b(1)
      q = a(0) - p2*b(0)
*
      b1(n-2) = 0
      b1(n-3) = -a(n)
      do 2 k=n-2,2,-1
         b1(k-2) = -b(k-1) - p1*b1(k-1) - p2*b1(k)
2     continue
      r1 = -b(0) - p1*b1(0) - p2*b1(1)
      q1 = -p2*b1(0)
*
      b2(n-2) = 0
      b2(n-3) = 0
      do 3 k=n-2,2,-1
         b2(k-2) = -p1*b2(k-1) - b(k) - p2*b2(k)
3     continue
      r2 = -p1*b2(0) - b(1) - p2*b2(1)
      q2 = -b(0) - p2*b2(0)
      end
*  ---------------------------------------------------------------------
      subroutine subsum ( m, n, x, f )
*  ---------------------------------------------------------------------
      implicit double precision (a-h,o-z)
      dimension x(n), f(m)
      common /toroot/ a(0:100), b(0:100), kco
      p1 = x(1)
      p2 = x(2)
      call syndiv(a,kco,p1,p2,b,r,q)
      f(1) = r
      f(2) = q
      end
*  ---------------------------------------------------------------------
      subroutine granal ( n, x, grad )
*  ---------------------------------------------------------------------
      implicit double precision (a-h,o-z)
      dimension x(n), grad(n)
      common /toroot/ a(0:100), b(0:100), kco
      dimension b1(0:100), b2(0:100)
      p1 = x(1)
      p2 = x(2)
      call  synder(a,kco,p1,p2,b,r,q,b1,b2,r1,r2,q1,q2)
      grad(1) = 2*(r*r1 + q*q1)
      grad(2) = 2*(r*r2 + q*q2)
      end
*  ---------------------------------------------------------------------
      subroutine janal ( m, n, x, fj, ld )
*  ---------------------------------------------------------------------
      implicit double precision (a-h,o-z)
      dimension x(n), fj(ld,n)
      common /toroot/ a(0:100), b(0:100), kco
      dimension b1(0:100), b2(0:100)
      p1 = x(1)
      p2 = x(2)
      call  synder(a,kco,p1,p2,b,r,q,b1,b2,r1,r2,q1,q2)
      fj(1,1) = r1
      fj(1,2) = r2
      fj(2,1) = q1
      fj(2,2) = q2
      end
